#include <sstream>
#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_image.h>
#include <math.h>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include "glm-0.9.2.6/glm/glm.hpp"
#include "glm-0.9.2.6/glm/gtc/matrix_transform.hpp"
#include "glm-0.9.2.6/glm/gtc/type_ptr.hpp"

#include "src/vector.h"
#include "src/keyboard.h"
#include "src/timer.h"
#include "src/matrix.h"
#include "src/glhelper.h"
#include "src/obj.h"

#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define CLAMP(x, y, z) ((x)<(y)?(y):((x)>(z)?(z):(x)))

// saves the texture int the portable pixmap format
void savePPM(unsigned char* buffer, int width, int height, int bpp, const char* image_out) {
	std::fstream of(image_out, std::ios_base::out | std::ios_base::trunc);
	of << "P3" << std::endl;
	of << width << " " << height << std::endl;
	of << "255" << std::endl;
	for (int j = 0; j < height; j++) {
		for (int i = 0; i < width; i++) {
			of << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 0] << " "
			   << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 1] << " "
			   << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 2] << " ";
		}
		of << std::endl;
	}
}

enum geometry  { NONE, PLANE, SPHERE };
enum materials { DIFFUSE, SPECULAR, REFRACTIVE };

// abstract base class: the plane and sphere objects are derived from this
class cObject {
    public:
	int material;
	vector3 color;
	vector3 emissive;
	int type;

	cObject();
	cObject(vector3 color, vector3 emissive, int material);
	~cObject();
	virtual void applyCamera(cMatrix& camera) = 0;
	virtual void save() = 0;
};

cObject::cObject() : color(vector3(1.0,1.0,1.0)), emissive(vector3(0.0,0.0,0.0)), material(DIFFUSE) { }
cObject::cObject(vector3 color, vector3 emissive, int material) : color(color), emissive(emissive), material(material) { }
cObject::~cObject() { std::cout << "object destroyed" << std::endl; }


class cPlane : public cObject {
    public:
	vector3 normal, normal_;
	vector3 point, point_;
	
	cPlane();
	cPlane(vector3 normal, vector3 point, vector3 color, vector3 emissive, int material);
	~cPlane();
	void applyCamera(cMatrix& camera);
	void save();
};

cPlane::cPlane() : normal(vector3(0.0,0.0,1.0)), point(vector3()), normal_(vector3(0.0,0.0,1.0)), point_(vector3()), cObject(vector3(1.0,1.0,1.0), vector3(0.0,0.0,0.0), material) { this->type = 1; }
cPlane::cPlane(vector3 normal, vector3 point, vector3 color, vector3 emissive, int material) : normal(normal), point(point), normal_(normal), point_(point), cObject(color, emissive, material) { this->type = 1; }
cPlane::~cPlane() { std::cout << "plane destroyed" << std::endl; }
void cPlane::applyCamera(cMatrix& camera) {
	vector3 p = this->point_ + this->normal_;
	this->point = camera.mult(this->point_);
	p = camera.mult(p);
	this->normal = (p - this->point).unit();
}
void cPlane::save() {
	this->normal_ = this->normal;
	this->point_ = this->point;
}


class cSphere : public cObject {
    public:
	vector3 center, center_;
	double radius, radius_;

	cSphere();
	cSphere(vector3 center, double radius, vector3 color, vector3 emissive, int material);
	~cSphere();
	void applyCamera(cMatrix& camera);
	void save();
};

cSphere::cSphere() : center(vector3(0.0,0.0,0.0)), center_(vector3(0.0,0.0,0.0)), radius(1.0), radius_(1.0), cObject(vector3(1.0,1.0,1.0), vector3(0.0,0.0,0.0), material) { this->type = 2; }
cSphere::cSphere(vector3 center, double radius, vector3 color, vector3 emissive, int material) : center(center), center_(center), radius(radius), radius_(radius), cObject(color, emissive, material) { this->type = 2; }
cSphere::~cSphere() { std::cout << "sphere destroyed" << std::endl; }
void cSphere::applyCamera(cMatrix& camera) {
	this->center = camera.mult(this->center_);
}
void cSphere::save() {
	this->center_ = this->center;
	this->radius_ = this->radius;
}

struct vertex_ {
	double x,  y,   z;
	double tx, ty, tz;
	double t[6];
};

class __vector {
  public:
	double x, y, z;

	__vector();
	__vector(double x, double y, double z);

	double dot(__vector w);
	double operator*(__vector w);
	__vector cross(__vector w);
	__vector add(__vector w);
	__vector sub(__vector w);
	__vector mult(double s);
	__vector h(__vector w);
	double length();
	__vector unit();
};

struct __object {
	int type;
	int material;
	double radius;

	__vector emission;
	__vector color;
	__vector normal;
	__vector point;
	__vector center;
};


//extern "C" {
	bool initializePathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object* _objects, __object*& current_objects_device, int object_count, float*& rand_device, int max_bounces, int width, int height);
	bool setupPathTracer(double*& current_doubles_device, __object*& current_objects_device, __object* _objects, int object_count, int width, int height);
	bool releasePathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object*& current_objects_device, float*& rand_device, int width, int height);
	bool runPathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, float*& rand_device, unsigned long long samples, __object objects[], __object*& current_objects_device, int object_count, int max_bounces, int width, int height);
//}


int main(int argc, char* argv[]) {

	int WIDTH = 640;
	int HEIGHT = 360;
	int BPP = 24;
//	int MAX_BOUNCES = 8;
	int MAX_BOUNCES = 5;
	
	cKeyboard kb;
	cTimer t0; double elapsed0;
  
	
	bool active = true;
	int samples = 0;


	SDL_Init(SDL_INIT_EVERYTHING);
	SDL_Surface *screen = mySDLInit(WIDTH, HEIGHT, BPP, false);
	SDL_Event event;

	
	std::vector<cObject*> objects;
// scene one
/*
	objects.push_back(new cPlane(vector3( 1, 0, 0), vector3(-456,   0,   0), vector3(0.75, 0.25, 0.25), vector3(), DIFFUSE)); // left
	objects.push_back(new cPlane(vector3(-1, 0, 0), vector3( 456,   0,   0), vector3(0.25, 0.25, 0.75), vector3(), DIFFUSE)); // right
	objects.push_back(new cPlane(vector3( 0, 0, 1), vector3(   0,   0,-456), vector3(0.75, 0.75, 0.75), vector3(), DIFFUSE)); // far
	objects.push_back(new cPlane(vector3( 0, 0,-1), vector3(   0,   0,2456), vector3(0.0,   0.0,  0.0), vector3(), DIFFUSE)); // near
	objects.push_back(new cPlane(vector3( 0,-1, 0), vector3(   0, 256,   0), vector3(0.75, 0.75, 0.75), vector3(), DIFFUSE)); // top
	objects.push_back(new cPlane(vector3( 0, 1, 0), vector3(   0,-256,   0), vector3(0.75, 0.75, 0.75), vector3(), DIFFUSE)); // bottom
	objects.push_back(new cSphere(vector3(-228,-106,-128), 150, vector3(1,1,1)*0.9, vector3(),         SPECULAR  ));
	objects.push_back(new cSphere(vector3( 228,-106, 0), 150, vector3(1,1,1)*0.999, vector3(),         REFRACTIVE));
//	objects.push_back(new cSphere(vector3( 356, 1024*4+240.0, -356), 1024*4, vector3(0,0,0),           vector3(5,5,5), DIFFUSE  )); // light source
	objects.push_back(new cSphere(vector3( 0, 1024*4+240.0, 0), 1024*4, vector3(0,0,0),           vector3(4,4,4), DIFFUSE  )); // light source
//	objects.push_back(new cSphere(vector3(-128, 256,  28), 128, vector3(0,0,0),           vector3(10,10,10), DIFFUSE  )); // light source
	{
		vector3  origin(-232,164,1455);

		vector3 forward = (vector3(0,0,0)-origin).unit();
		vector3    up(0, 1,  0);
		vector3 right = (forward.cross(up)).unit();
			  up = (right.cross(forward)).unit();

		double cam[] = {  right.x,   right.y,   right.z, -(right*origin),
				     up.x,      up.y,      up.z, -(up*origin),
				forward.x, forward.y, forward.z, -(forward*origin),
					0,         0,         0,         1 };

		cMatrix camera(4, 4, cam);
		//camera.output();

		for (int i = 0; i < objects.size(); i++) {
			objects[i]->applyCamera(camera);
			objects[i]->save();
		}
	}
*/

// scene two
	double a = 0.9;
	double b = 0.8;
	objects.push_back(new cPlane(vector3( 0,-1, 0), vector3(   0, 8192,   0), vector3(0,0,0), vector3(1,1,1), DIFFUSE)); // top
	objects.push_back(new cPlane(vector3( 0, 1, 0), vector3(   0, 0,   0), vector3(b,b,b), vector3(), DIFFUSE)); // bottom
	for (int i = 0; i < 10; i++) {
//	for (int i = 0; i < 50; i++) {
		double sx = rand()/(double)RAND_MAX;
		double sy = rand()/(double)RAND_MAX;
		double sz = rand()/(double)RAND_MAX;
		double sr = rand()/(double)RAND_MAX;
		sx -= 0.5; sz -= 0.5; sr += 0.5;
		sx *= 1000.0*WIDTH/HEIGHT; sy *= 200; sz *= 1000; sr *= 80;
		int type = rand()%2;
		//a = type == 2 ? 1 : 0.9;
		objects.push_back(new cSphere(vector3(sx,sy,sz), sr, vector3(a,a,a), vector3(), type  )); // light source
	}

	{
		vector3  origin(0,1600,1000);

		vector3 forward = (vector3(0,0,0)-origin).unit();
		vector3    up(0, 1,  0);
		vector3 right = (forward.cross(up)).unit();
			  up = (right.cross(forward)).unit();

		double cam[] = {  right.x,   right.y,   right.z, -(right*origin),
				     up.x,      up.y,      up.z, -(up*origin),
				forward.x, forward.y, forward.z, -(forward*origin),
					0,         0,         0,         1 };

		cMatrix camera(4, 4, cam);
		//camera.output();

		for (int i = 0; i < objects.size(); i++) {
			objects[i]->applyCamera(camera);
			objects[i]->save();
		}
	}

	__object _objects[objects.size()];

	unsigned char *current_buffer_host;
	unsigned char *current_buffer_device;
	double *current_doubles_device;
	__object *current_objects_device;
	float *rand_device;
	initializePathTracer(current_buffer_host, current_buffer_device, current_doubles_device, _objects, current_objects_device, objects.size(), rand_device, MAX_BOUNCES, WIDTH, HEIGHT);


	// shaders
	GLuint glProgram, glShaderV, glShaderF;
	GLint vertex, texCoord, Projection, View, Model, textureSample;
	createProgram(glProgram, glShaderV, glShaderF, "src/vertex.sh", "src/fragment.sh");
	vertex        = glGetAttribLocation(glProgram, "vertex");
	texCoord      = glGetAttribLocation(glProgram, "texCoord");
	Projection    = glGetUniformLocation(glProgram, "Projection");
	View          = glGetUniformLocation(glProgram, "View");
	Model         = glGetUniformLocation(glProgram, "Model");
	textureSample = glGetUniformLocation(glProgram, "textureSample");

	// vertices
	const int vertex_count = 4;
	vertex_ vertices[vertex_count];
	vertices[0].x  = 0;     vertices[0].y  = 0;      vertices[0].z = 0;     vertices[0].tx = 0;     vertices[0].ty = 0;	
	vertices[1].x  = WIDTH; vertices[1].y  = 0;      vertices[1].z = 0;     vertices[1].tx = 1;     vertices[1].ty = 0;	
	vertices[2].x  = WIDTH; vertices[2].y  = HEIGHT; vertices[2].z = 0;     vertices[2].tx = 1;     vertices[2].ty = 1;
	vertices[3].x  = 0;     vertices[3].y  = HEIGHT; vertices[3].z = 0;     vertices[3].tx = 0;     vertices[3].ty = 1;
	GLuint vbo_vertices;
	glGenBuffers(1, &vbo_vertices);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertex_) * vertex_count, vertices, GL_DYNAMIC_DRAW);

	// indices
	const int indices_count = 6;
	unsigned int indices[indices_count] = {0, 1, 2, 2, 3, 0};
	GLuint vbo_indices;
	glGenBuffers(1, &vbo_indices);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int) * indices_count, indices, GL_STATIC_DRAW);

	// create a texture for rendering the result
	GLuint texture;
	setupTexture(texture);

	
	glm::mat4 projection = glm::ortho(0.0f, (float)WIDTH, (float)HEIGHT, 0.0f, -5.0f, 5.0f); 
	glm::mat4 view       = glm::mat4(1.0f);
	glm::mat4 model      = glm::mat4(1.0f);

	bool processed = false;
	double x = 0, z = 0, theta = M_PI/2;
	while (active) {

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			    case SDL_QUIT: active = false; break;
			}
		}

		elapsed0 = t0.elapsed(true);

		if (kb.getKeyState(KEY_UP))    { processed = false; z += 40.0f*elapsed0*sin(theta); x += 20.f*elapsed0*cos(theta); }
		if (kb.getKeyState(KEY_DOWN))  { processed = false; z -= 40.0f*elapsed0*sin(theta); x -= 20.f*elapsed0*cos(theta); }
		if (kb.getKeyState(KEY_LEFT))  { processed = false; theta += 1.0f*elapsed0; }
		if (kb.getKeyState(KEY_RIGHT)) { processed = false; theta -= 1.0f*elapsed0; }

		if (!processed) {
			vector3  origin(x, 0, z);

			vector3 forward = vector3(cos(theta), 0, sin(theta)).unit();
			vector3 up      = vector3(0, 1, 0);
			vector3 right   = up.cross(forward).unit();

			double cam[] = {  right.x,   right.y,   right.z, -(right*origin),
					     up.x,      up.y,      up.z, -(up*origin),
					forward.x, forward.y, forward.z, -(forward*origin),
						0,         0,         0,         1 };

			cMatrix camera(4, 4, cam);
		//	camera.output();

			for (int i = 0; i < objects.size(); i++) objects[i]->applyCamera(camera);

			for (int i = 0; i < objects.size(); i++) {

				_objects[i].type = objects[i]->type;
				_objects[i].material = objects[i]->material;

				_objects[i].color.x = objects[i]->color.x;
				_objects[i].color.y = objects[i]->color.y;
				_objects[i].color.z = objects[i]->color.z;

				_objects[i].emission.x = objects[i]->emissive.x;
				_objects[i].emission.y = objects[i]->emissive.y;
				_objects[i].emission.z = objects[i]->emissive.z;

				if (objects[i]->type == 1) {
					_objects[i].normal.x = ((cPlane *)objects[i])->normal.x;
					_objects[i].normal.y = ((cPlane *)objects[i])->normal.y;
					_objects[i].normal.z = ((cPlane *)objects[i])->normal.z;
					_objects[i].point.x = ((cPlane *)objects[i])->point.x;
					_objects[i].point.y = ((cPlane *)objects[i])->point.y;
					_objects[i].point.z = ((cPlane *)objects[i])->point.z;
				} else {
					_objects[i].center.x = ((cSphere *)objects[i])->center.x;
					_objects[i].center.y = ((cSphere *)objects[i])->center.y;
					_objects[i].center.z = ((cSphere *)objects[i])->center.z;
					_objects[i].radius = ((cSphere *)objects[i])->radius;
				}
			}
			processed = true;
			samples = 0;
			setupPathTracer(current_doubles_device, current_objects_device, _objects, objects.size(), WIDTH, HEIGHT);
		}

		samples++;

		samples = runPathTracer(current_buffer_host, current_buffer_device, current_doubles_device, rand_device, samples, _objects, current_objects_device, objects.size(), MAX_BOUNCES, WIDTH, HEIGHT);
		std::cout << "samples: " << samples << std::endl;

if (samples == 50000) active = false;

		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

		glUseProgram(glProgram);

		// texture
		glUniform1i(textureSample, 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, WIDTH, HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, current_buffer_host);

		// Project, View, and Model uniforms
		glUniformMatrix4fv(Projection, 1, GL_FALSE, glm::value_ptr(projection));
		glUniformMatrix4fv(View, 1, GL_FALSE, glm::value_ptr(view));
		model = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f,0.0f,-2.0f));
		glUniformMatrix4fv(Model, 1, GL_FALSE, glm::value_ptr(model));

		// vertex and texture attributes
		glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
		glEnableVertexAttribArray(vertex);
		glVertexAttribPointer(vertex, 3, GL_DOUBLE, GL_FALSE, sizeof(vertex_), 0);
		glEnableVertexAttribArray(texture);
		glVertexAttribPointer(texCoord, 2, GL_DOUBLE, GL_FALSE, sizeof(vertex_), (char *)NULL + 24);

		// draw it
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
		glDrawElements(GL_TRIANGLES, indices_count, GL_UNSIGNED_INT, 0);
		SDL_GL_SwapBuffers();
	}

	savePPM(current_buffer_host, WIDTH, HEIGHT, BPP, "screen.ppm");
	
	releasePathTracer(current_buffer_host, current_buffer_device, current_doubles_device, current_objects_device, rand_device, WIDTH, HEIGHT);

	// release the objects
	for (int i = 0; i < objects.size(); i++) delete objects[i];
	
	// release the texture, vertex buffer objects, and shader program.. also shut down SDL
	deleteTexture(texture);
	glDeleteBuffers(1, &vbo_indices);
	glDeleteBuffers(1, &vbo_vertices);
	releaseProgram(glProgram, glShaderV, glShaderF);
	SDL_Quit();

	return 0;
}